/*
    A wrapper for mplayer that keeps videos between two networked hosts synced.
    Copyright (C) 2015  alicia@ion.nu

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <unistd.h>
#include <stdio.h>
#include <poll.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netdb.h>
#include <signal.h>

int main(int argc, char** argv)
{
  signal(SIGPIPE, SIG_IGN);
  signal(SIGCHLD, exit); // If we exit mplayer we should exit too
  int myargc=1;
  char** myargv=malloc(sizeof(char*));
  myargv[0]=argv[0];
  char* client=0;
  char* server=0;
  char* file=0;
  // TODO: some kind of NAT traversal?
  int i;
  for(i=1; i<argc; ++i)
  {
    if(!strcmp(argv[i], "-server"))
    {
      if(i>argc){printf("-server requires an argument\n"); return 1;}
      ++i;
      server=argv[i];
      continue;
    }else
    if(!strcmp(argv[i], "-client"))
    {
      if(i>argc){printf("-client requires an argument\n"); return 1;}
      ++i;
      client=argv[i];
      continue;
    }else
    if(argv[i][0]!='-' && !access(argv[i], R_OK)) // Grab the filename. We would have checked for the first argument which isn't an argument of a flag, but there are over 200 flags with arguments that we would have to check for.
    {
      file=argv[i];
    }
    ++myargc;
    myargv=realloc(myargv, sizeof(char*)*myargc);
    myargv[myargc-1]=argv[i];
  }
  if((client && server)||!(client || server))
  {
    printf("Usage:\n"
    "%s [-client <server>:<port>] [-server <port>] <options which will be passed to mplayer>\n"
    "The -client and -server options are mutually exclusive and mandatory.\n"
    "\n"
    "-client <server>:<port>  Act as the client, connect to a remote host.\n"
    "-server <port>           Act as the server, wait for connections on <port>\n"
    "For options related to playback, see mplayer's --help text and documentation\n", argv[0]);
    return 1;
  }
  myargc+=2;
  myargv=realloc(myargv, sizeof(char*)*(myargc+1));
  myargv[myargc-2]=strdup("-slave");
  myargv[myargc-1]=strdup("-identify");
  myargv[myargc]=0;
  // Establish connection with client/server
  struct addrinfo* ai;
  char* port;
  if(server)
  {
    port=server;
  }else{
    port=strchr(client, ':');
    if(!port){printf("No port specified\n"); return 1;}
    port[0]=0;
    port=&port[1];
  }
  if(getaddrinfo(client?client:"0.0.0.0", port, 0, &ai))
  {
    printf("Failed to resolve '%s'\n", client?client:server);
    return 1;
  }
  int other=socket(ai->ai_family, SOCK_STREAM, IPPROTO_TCP);
  int other_listen;
  if(client)
  {
    while(1)
    {
      printf("Trying to connect with partner...\n");
      if(!connect(other, ai->ai_addr, ai->ai_addrlen)){break;}
      perror("connect");
      sleep(1);
    }
  }else{
    printf("Waiting for partner to connect... ");
    fflush(stdout);
    other_listen=other;
    while(bind(other_listen, ai->ai_addr, ai->ai_addrlen)){perror("bind");sleep(5);}
    listen(other_listen, 1);
    struct sockaddr addr;
    socklen_t addrlen=sizeof(addr);
    other=accept(other_listen, &addr, &addrlen);
    printf("done\n");
  }
  // TODO: Instead of just comparing filenames, compare checksums
  if(file)
  {
    char* slash=strrchr(file, '/');
    if(slash){file=&slash[1];}
    dprintf(other, "file:%s\n", file);
  }
  int in[2];
  int out[2];
  pipe(in);
  pipe(out);
  if(!fork())
  {
    close(other);
    if(server){close(other_listen);}
    dup2(in[0], 0);
    dup2(out[1], 1);
    close(in[1]);
    close(out[0]);
    execvp("mplayer", myargv);
    _exit(1);
  }
  close(in[0]);
  close(out[1]);
  struct pollfd pfd[3];
  pfd[0].fd=out[1];
  pfd[0].events=POLLIN;
  pfd[0].revents=0;
  pfd[1].fd=other;
  pfd[1].events=POLLIN;
  pfd[1].revents=0;
  pfd[2].fd=0;
  pfd[2].events=POLLIN;
  pfd[2].revents=0;
  char buf[1024];
  int position=0;
  time_t lastsend=time(0);
  char first=1;
  while(1)
  {
    poll(pfd, 3, -1);
    if(pfd[0].revents) // Got mplayer output
    {
      pfd[0].revents=0;
      size_t len=read(out[0], buf, 1023);
      if(len<1){break;}
      write(1, buf, len);
      buf[len]=0;
      char* pos=buf;
      if(!strncmp(pos, "A:", 2)||!strncmp(pos, "V:", 2))
      {
        for(pos=&pos[2]; pos[0]==' '; pos=&pos[1]);
        position=atoi(pos);
        time_t now=time(0);
        if(now>lastsend+5)
        {
// printf("Sending position %i\n", position);
          lastsend=now;
          dprintf(other, "%i\n", position);
        }
      }
    }
    if(pfd[1].revents) // Got a timestamp from partner
    {
      pfd[1].revents=0;
      size_t len=read(other, buf, 1023);
      if(len<1) // Lost connection, reconnect/accept new connection
      {
        close(other);
        if(client)
        {
          close(other);
          other=socket(ai->ai_family, SOCK_STREAM, IPPROTO_TCP);
          pfd[1].fd=other;
          while(1)
          {
            fprintf(stderr, "Trying to reconnect with partner...\n");
            if(!connect(other, ai->ai_addr, ai->ai_addrlen)){break;}
            perror("connect");
            sleep(1);
          }
          continue;
        }else{
          fprintf(stderr, "Waiting for partner to reconnect... ");
          fflush(stdout);
          struct sockaddr addr;
          socklen_t addrlen=sizeof(addr);
          other=accept(other_listen, &addr, &addrlen);
          printf("done\n");
          continue;
        }
      }
      buf[len]=0;
      if(!strncmp(buf, "file:", 5)) // Check that filenames match
      {
        char* eol;
        if((eol=strchr(buf, '\n'))){eol[0]=0;}
        if(file && strcmp(file, &buf[5]))
        {
          fprintf(stderr, "\nLocal ('%s') and remote ('%s') filenames seem to differ, are you watching the same thing?\n\n", file, &buf[5]);
        }
        continue;
      }
      int pos=atoi(buf);
      if(position>pos+2 || first) // Go back to stay in sync
      {
        fprintf(stderr, "Syncing! (%i)\n", pos);
        dprintf(in[1], "seek %i 2\n", pos);
        position=pos;
      }
      first=0;
    }
    if(pfd[2].revents) // Got input
    {
      pfd[2].revents=0;
      size_t len=read(0, buf, 1024);
      if(len<1){break;}
      write(in[1], buf, len);
    }
  }
  return 0;
}
